@@ -87,70 +87,85 @@ The path can be absolute or relative to that of the linted file.")
 This is bound dynamically while the checks run.")
 
 (defconst package-lint-backport-libraries
-  (list (cons 'cl-lib "\\`cl-")
-        (cons 'cl-generic "\\`cl-\\(?:def\\)?generic")
-        (cons 'cl-print "\\`cl-print")
-        (cons 'map "\\`map-")
-        (cons 'nadvice "\\`advice-")
-        (cons 'seq "\\`seq-")
-        (cons 'let-alist "\\`let-alist"))
-  "A sequence of (FEATURE . SYMBOL-NAME-MATCH) for backport libraries.
+  (list (list 'cl-lib "\\`cl-")
+        (list 'cl-generic "\\`cl-\\(?:def\\)?generic")
+        (list 'cl-print "\\`cl-print")
+        (list 'jsonrpc "\\`jsonrpc-")
+        (list 'map "\\`map-")
+        (list 'nadvice "\\`advice-")
+        (list 'ol "\\`org-link-" 'org "9.3")
+        (list 'seq "\\`seq-")
+        (list 'let-alist "\\`let-alist"))
+  "A sequence of (FEATURE SYMBOL-NAME-MATCH) for backport libraries.
 These are libraries that are built into newer Emacsen and also
-published in ELPA for use by older Emacsen.")
+published in ELPA for use by older Emacsen.
 
-(eval-and-compile
-  (defun package-lint--match-symbols (symbols)
-    "Return a predicate which take a symbol and reports whether it is among `SYMBOLS'."
-    (let ((tbl (make-hash-table)))
-      (dolist (s symbols)
-        (puthash s t tbl))
-      (lambda (v) (gethash v tbl))))
+Alternatively, items can take a form of (FEATURE
+SYMBOL-NAME-MATCH PACKAGE VERSION?) where PACKAGE is the name of
+a backport library shipping the feature and VERSION is an
+optional minimum version containing the feature.")
 
-  (let ((stdlib-changes (with-temp-buffer
-                          (insert-file-contents
-                           (expand-file-name "data/stdlib-changes"
-                                             (if load-file-name
-                                                 (file-name-directory load-file-name)
-                                               default-directory)))
-                          (read (current-buffer)))))
+(defconst package-lint-symbol-info
+  (let* ((stdlib-changes (with-temp-buffer
+                           (insert-file-contents
+                            (expand-file-name "data/stdlib-changes"
+                                              (if load-file-name
+                                                  (file-name-directory load-file-name)
+                                                default-directory)))
+                           (read (current-buffer))))
+         (info (make-hash-table)))
+    (pcase-dolist (`(,version . ,data) stdlib-changes)
+      (pcase-dolist (`(,syms . ,action)
+                     (let-alist data
+                       (list (cons .variables.added 'variable-added)
+                             (cons .variables.removed 'variable-removed)
+                             (cons .features.added 'library-added)
+                             (cons .features.removed 'library-removed)
+                             (cons .functions.added 'function-added)
+                             (cons .functions.removed 'function-removed))))
+        (dolist (sym syms)
+          (puthash sym (cons (cons action version) (gethash sym info)) info))))
+    info)
+  "A hash table from SYMBOL to a list of events in its history.
+Each event is of the form (ACTION . EMACS-VER), where ACTION is a
+symbol such as 'variable-added.")
 
-    (defconst package-lint--libraries-added-alist
-      (mapcar (lambda (version-data)
-                (let ((version (car version-data))
-                      (added-libraries (let-alist (cdr version-data) .features.added)))
-                  (cons version (package-lint--match-symbols added-libraries))))
-              stdlib-changes)
-      "An alist of library names and when they were added to Emacs.")
+(defun package-lint-symbol-info (sym)
+  "Retrieve information about SYM, as an alist of (ACTION . EMACS-VER)."
+  (gethash (cl-etypecase sym
+             (string (intern sym))
+             (symbol sym))
+           package-lint-symbol-info))
 
-    (defconst package-lint--libraries-removed-alist
-      (mapcar (lambda (version-data)
-                (let ((version (car version-data))
-                      (removed-libraries (let-alist (cdr version-data) .features.removed)))
-                  (cons version (package-lint--match-symbols removed-libraries))))
-              stdlib-changes)
-      "An alist of library names and when they were added to Emacs.")
+(defun package-lint--read-known-symbol (prompt)
+  "Read a symbol `package-lint-symbol-info' with PROMPT."
+  (let* ((at-point (symbol-at-point))
+         (s (completing-read
+             prompt
+             (cl-loop for k being the hash-keys of package-lint-symbol-info collect k)
+             nil
+             t
+             (when at-point (symbol-name at-point)))))
+    (when s (intern s))))
 
-    (defconst package-lint--functions-and-macros-added-alist
-      (mapcar (lambda (version-data)
-                (let ((version (car version-data))
-                      (added-functions (let-alist (cdr version-data) .functions.added)))
-                  (cons version (package-lint--match-symbols added-functions))))
-              stdlib-changes)
-      "An alist of function/macro names and when they were added to Emacs.")
+;;;###autoload
+(defun package-lint-describe-symbol-history (sym)
+  "Show the version history of SYM, if any."
+  (interactive (list (package-lint--read-known-symbol "Show history of symbol: ")))
+  (message
+   (concat (format "History of %s:\n" sym)
+           (mapconcat (lambda (info)
+                        (pcase-let ((`(,action . ,version) info))
+                          (format "* %s in Emacs %s"
+                                  action (mapconcat 'number-to-string version "."))))
+                      (sort (package-lint-symbol-info sym)
+                            (lambda (a b) (version-list-<= (cdr a) (cdr b))))
+                      "\n"))))
 
-    (defconst package-lint--functions-and-macros-removed-alist
-      (mapcar (lambda (version-data)
-                (let ((version (car version-data))
-                      (removed-functions (let-alist (cdr version-data) .functions.removed)))
-                  (cons version (package-lint--match-symbols removed-functions))))
-              stdlib-changes)
-      "An alist of function/macro names and when they were removed from Emacs.")
-
-    (defun package-lint--added-or-removed-function-p (sym)
-      "Predicate that returns t if SYM is a function added/removed in any known emacs version."
-      (cl-some (lambda (x) (funcall (cdr x) sym))
-               (append package-lint--functions-and-macros-added-alist
-                       package-lint--functions-and-macros-removed-alist)))))
+(defun package-lint--added-or-removed-function-p (sym)
+  "Predicate that returns t if SYM is a function added/removed in any known emacs version."
+  (let-alist (package-lint-symbol-info sym)
+    (or .function-added .function-removed)))
 
 (defconst package-lint--sane-prefixes
   (rx
@@ -462,7 +477,7 @@ required version PACKAGE-VERSION.  If not, raise an error for DEP-POS."
          'error
          "You can only depend on Emacs version 24 or greater: package.el for Emacs 23 does not support the \"emacs\" pseudopackage."
          dep-pos))
-       ((version-list-<= '(27) package-version)
+       ((version-list-<= '(28) package-version)
         (package-lint--error-at-point
          'warning
          "This makes the package uninstallable in all released Emacs versions."
@@ -520,7 +535,7 @@ required version PACKAGE-VERSION.  If not, raise an error for DEP-POS."
   (save-excursion
     (save-match-data
       (and (re-search-backward
-            (concat "(fboundp\\s-+'" (regexp-quote sym) "\\_>") (point-min) t)
+            (concat "(" (regexp-opt '("fboundp" "macrop")) "\\s-+#?'" (regexp-quote sym) "\\_>") (point-min) t)
            (not (package-lint--inside-comment-or-string-p))))))
 
 (defun package-lint--map-regexp-match (regexp callback)
@@ -546,39 +561,54 @@ CALLBACK."
               (unless (package-lint--inside-comment-or-string-p)
                 (apply #'package-lint--error-at-point err)))))))))
 
-(defun package-lint--check-version-regexp-list (valid-deps list symbol-regexp type)
+(defun package-lint--check-version-regexp-list (valid-deps symbol-regexp type)
   "Warn if symbols matched by SYMBOL-REGEXP are unavailable in the target Emacs.
 The target Emacs version is taken from VALID-DEPS, which are the
-declared dependencies of this package.  LIST is an alist
-of (VERSION . PRED), where PRED is passed the sym.  TYPE is the
+declared dependencies of this package.  TYPE is the
 type of the symbol, either FUNCTION or FEATURE."
   (let ((emacs-version-dep (or (cadr (assq 'emacs valid-deps)) '(0))))
-    (pcase-dolist (`(,added-in-version . ,pred) list)
-      (when (version-list-< emacs-version-dep added-in-version)
-        (package-lint--map-regexp-match
-         symbol-regexp
-         (lambda (sym)
-           (when (funcall pred (intern sym))
+    (package-lint--map-regexp-match
+     symbol-regexp
+     (lambda (sym)
+       (let-alist (package-lint-symbol-info sym)
+         (let ((added-in-version (cl-ecase type
+                                   ('function .function-added)
+                                   ('feature .library-added))))
+           (when (and added-in-version (version-list-< emacs-version-dep added-in-version))
              (unless (and (eq type 'function) (package-lint--seen-fboundp-check-for sym))
-               (let ((available-backport
-                      (cond
-                       ((eq type 'feature)
-                        (cl-some (lambda (bp)
-                                   (when (string= (car bp) sym)
-                                     (car bp)))
-                                 package-lint-backport-libraries))
-                       ((eq type 'function)
-                        (cl-some (lambda (bp)
-                                   (when (string-match-p (cdr bp) sym)
-                                     (car bp)))
-                                 package-lint-backport-libraries)))))
-                 (unless (and available-backport (assoc available-backport valid-deps))
+               (let* ((available-backport-with-ver
+                       (cl-ecase type
+                         ('feature
+                          (cl-some (lambda (bp)
+                                     (when (string= (car bp) sym)
+                                       (or (cddr bp)
+                                           (list (car bp)))))
+                                   package-lint-backport-libraries))
+                         ('function
+                          (cl-some (lambda (bp)
+                                     (when (string-match-p (nth 1 bp) sym)
+                                       (or (cddr bp)
+                                           (list (car bp)))))
+                                   package-lint-backport-libraries))))
+                      (available-backport (car available-backport-with-ver))
+                      (required-backport-version (cadr available-backport-with-ver))
+                      (matching-dep (when available-backport
+                                      (assoc available-backport valid-deps))))
+                 (unless (and matching-dep
+                              (or (not required-backport-version)
+                                  (version-list-<= (version-to-list required-backport-version)
+                                                   (cadr matching-dep))))
                    (list
                     'error
                     (format "You should depend on (emacs \"%s\")%s if you need `%s'."
                             (mapconcat #'number-to-string added-in-version ".")
                             (if available-backport
-                                (format " or the %s package" available-backport)
+                                (format " or the %s package"
+                                        (if required-backport-version
+                                            (format "(%s \"%s\")"
+                                                    available-backport
+                                                    required-backport-version)
+                                          available-backport))
                               "")
                             sym))))))))))))
 
@@ -620,7 +650,6 @@ type of the symbol, either FUNCTION or FEATURE."
   "Warn about use of libraries that are not available in the Emacs version in VALID-DEPS."
   (package-lint--check-version-regexp-list
    valid-deps
-   package-lint--libraries-added-alist
    package-lint--unconditional-require-regexp
    'feature))
 
@@ -629,14 +658,13 @@ type of the symbol, either FUNCTION or FEATURE."
   (package-lint--map-regexp-match
    package-lint--unconditional-require-regexp
    (lambda (sym)
-     (cl-block return
-       (pcase-dolist (`(,removed-in-version . ,pred) package-lint--libraries-removed-alist)
-         (when (funcall pred (intern sym))
-           (cl-return-from return
-             (list
-              'error
-              (format "The `%s' library was removed in Emacs version %s."
-                      sym (mapconcat #'number-to-string removed-in-version "."))))))))))
+     (let-alist (package-lint-symbol-info sym)
+       (let ((removed-in-version .library-removed))
+         (when removed-in-version
+           (list
+            'error
+            (format "The `%s' library was removed in Emacs version %s."
+                    sym (mapconcat #'number-to-string removed-in-version ".")))))))))
 
 (defconst package-lint--function-name-regexp
   (rx
@@ -657,7 +685,6 @@ type of the symbol, either FUNCTION or FEATURE."
   "Warn about use of functions/macros that are not available in the Emacs version in VALID-DEPS."
   (package-lint--check-version-regexp-list
    valid-deps
-   package-lint--functions-and-macros-added-alist
    package-lint--function-name-regexp
    'function))
 
@@ -669,20 +696,15 @@ the Emacs dependency matches the re-addition."
   (package-lint--map-regexp-match
    package-lint--function-name-regexp
    (lambda (sym)
-     (cl-block return
-       (pcase-dolist (`(,removed-in-version . ,pred) package-lint--functions-and-macros-removed-alist)
-         (when (funcall pred (intern sym))
+     (let-alist (package-lint-symbol-info sym)
+       (let ((removed-in-version .function-removed))
+         (when removed-in-version
            (let ((emacs-version-dep (or (cadr (assq 'emacs valid-deps)) '(0))))
-             (unless (cl-some (lambda (dep)
-                                (pcase-let ((`(,ver . ,pred) dep))
-                                  (and (version-list-<= ver emacs-version-dep)
-                                       (funcall pred (intern sym)))))
-                              package-lint--functions-and-macros-added-alist)
-               (cl-return-from return
-                 (list
-                  'error
-                  (format "`%s' was removed in Emacs version %s."
-                          sym (mapconcat #'number-to-string removed-in-version "."))))))))))))
+             (unless (and .function-added (version-list-<= .function-added emacs-version-dep))
+               (list
+                'error
+                (format "`%s' was removed in Emacs version %s."
+                        sym (mapconcat #'number-to-string removed-in-version ".")))))))))))
 
 (defun package-lint--check-lexical-binding-is-on-first-line ()
   "Check that any `lexical-binding' declaration is on the first line of the file."
@@ -815,12 +837,19 @@ DESC is a struct as returned by `package-buffer-info'."
 (defun package-lint--check-provide-form (desc)
   "Check the provide form for package with descriptor DESC.
 DESC is a struct as returned by `package-buffer-info'."
-  (let ((name (package-lint--package-desc-name desc))
-        (feature (package-lint--provided-feature)))
-    (unless (string-equal (symbol-name name) feature)
-      (package-lint--error-at-bob
-       'error
-       (format "There is no (provide '%s) form." name)))))
+  (let ((name (symbol-name (package-lint--package-desc-name desc))))
+    (if (string-match-p "-theme\\'" name)
+        (let ((theme-name (replace-regexp-in-string "-theme\\'" "" name))
+              (provided-theme (package-lint--provided-theme)))
+          (unless (string-equal theme-name provided-theme)
+            (package-lint--error-at-bob
+             'error
+             (format "There is no (provide-theme '%s) form." theme-name))))
+      (let ((feature (package-lint--provided-feature)))
+        (unless (string-equal name feature)
+          (package-lint--error-at-bob
+           'error
+           (format "There is no (provide '%s) form." name)))))))
 
 (defun package-lint--check-provide-form-secondary-file ()
   "Check there is a provide form."
@@ -892,7 +921,7 @@ Valid definition names are:
   "Verify that symbol DEFINITIONS start with package PREFIX."
   (pcase-dolist (`(,name . ,position) definitions)
     (unless (package-lint--valid-definition-name-p name prefix position)
-      (if (package-lint--added-or-removed-function-p (intern name))
+      (if (package-lint--added-or-removed-function-p name)
           (package-lint--error-at-point
            'error
            (format "Define compatibility functions with a prefix, e.g. \"%s--%s\", and use `defalias' where they exist."
@@ -1104,7 +1133,7 @@ The returned list is of the form (SYMBOL-NAME . POSITION)."
       (pcase entry
         ((and `(,submenu-name . ,submenu-elements)
               (guard (consp submenu-elements)))
-         (when (member submenu-name '("Variables" "Defuns"))
+         (when (member submenu-name '("Variables" "Defuns" "Types"))
            (setq result (nconc (reverse submenu-elements) result))))
         (_
          (push entry result))))
@@ -1124,6 +1153,13 @@ The returned list is of the form (SYMBOL-NAME . POSITION)."
           ((re-search-backward "(provide-me)" nil t)
            (file-name-sans-extension (file-name-nondirectory buffer-file-name))))))
 
+(defun package-lint--provided-theme ()
+  "Return the last-provided theme name, as a string, or nil if none."
+  (save-excursion
+    (goto-char (point-max))
+    (when (re-search-backward (rx "(provide-theme '" (group (1+ (or (syntax word) (syntax symbol))))) nil t)
+      (match-string-no-properties 1))))
+
 (defun package-lint--get-package-prefix ()
   "Return package prefix string (i.e. the symbol the package `provide's).
 Prefix is returned without any `-mode' suffix."
@@ -1245,8 +1281,7 @@ whether or not warnings alone produce a non-zero exit code."
             (widen)
             (goto-char (point-min))
             (re-search-forward
-             (concat lm-header-prefix
-                     (rx (or "Version" "Package-Version" "Package-Requires")))
+             (lm-get-header-re (rx (or "Version" "Package-Version" "Package-Requires")))
              nil t))))))
 
 (provide 'package-lint)
